#!/usr/bin/python3

#
#        MiniQuant --  Wrapper for PicoQuant's HydraHarp access libraries.
#        Copyright (C) 2022 Florin Boariu.
#
#        This program is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        This program is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

#
# Unit-testing -- it is expected to have a Harp device attached if this
# is to succeed.
#

import pytest
from os import environ as env

if 'MINIQUANT_TEST' in env and env['MINIQUANT_TEST'] != 0:
    from miniquant import mock_harp as harp
else:
    from miniquant import harp

def test_library():
    ##    Finding / initializing the HH-library and singleton wrapper

    if harp.monster_binding is not None:
        
        # This should deliver a valid m, which is the same as monster_library        
        m = harp.Monster()
        assert (m is harp.monster_binding)

        assert len(m.libraryVersion) > 1
        
    else:
        
        # this is the same as default initialisation -- it should fail
        with pytest.raises(Exception):
            m = harp.Monster()

        # special env-var for testing with the *real* location of the library :-)
        if env.get("TEST_MINIQUANT_HHLIB"):
            m = harp.Monster(env.get("TEST_MINIQUANT_HHLIB"))

            assert (m.hlib is not None)

            assert len(m.libraryVersion) > 1


def test_findharp():
    ## Finding a harp device

    m = harp.Monster()
    h = m.find("first")
    assert (h is not None)

    h2 = m.find(h.serial)
    assert (h2.serial == h.serial)

    ## FIXME: at this point we've opened a device twice (h.serial)
    ## What does happen if we closed one? What is supposed to happen?!...
    ## need to test thath.
    

def test_notfindharp():
    mon = harp.Monster()
    with pytest.raises(OSError):
        h3 = mon.find("unlikely")

        
def test_findall():
    h4 = harp.Monster().find("all")
    assert(len(h4) > 0)

@pytest.fixture
def harp_device():
    h = harp.Monster().find('any')

    print ("Initializing Harp No.", h.serial, "...")
    h.initialize(harp.MeasurementMode.Continuous,
                 harp.ReferenceClock.Internal)
    print ("Initialization finished.")
    
    return h

def test_api_getters(harp_device):
    '''
    Tests all kinds of API "HH_Get..." wrappers.
    '''

    hi = harp_device.hardwareInfo
    print ("Hardware Info:", hi)
    assert (len(hi) == 3)

    sn = harp_device.serialNumber
    print ("Serial Number:", sn)
    assert (sn == harp_device.serial)

    ft = harp_device.features
    print ("Features: ")
    [ print("  -", f) for f in ft ]
    
    res = harp_device.baseResolution
    print ("Base Resolution:", res)
    assert (len(res) == 2)

    mods = harp_device.modules
    assert (len(mods) > 0)
    print ("Installed Modules:", len(mods))
    print ("Module Info:")
    [ print ("  -", m.info) for m in mods ]

    ichs = harp_device.channels
    assert (len(ichs) > 0) ## only true for the HydraHarp
    print ("Installed Input Channels:", len(ichs))
    print ("Channel info:")
    [ print("  - module index:", c.moduleIndex,
            "\n    enabled:", c.enabled,
            "\n    offset:", c.offset,
            "\n    cfd:", c.cfd,
            "\n    count rate:", c.countRate) for c in ichs ]

    flg = harp_device.flags
    print("Flags:")
    [ print("  -", f) for f in flg ]

    wrn = harp_device.warnings
    print("Warnings:")
    [ print("  -", w) for w in wrn ]

    hdbg = harp_device.hardwareDebugInfo
    print ("H/W Debug Info: \n---cut-here---")
    print (hdbg, "\n---/cut-here---")

    print ("Sync channel:")
    print ("  - div:", harp_device.syncDiv)
    print ("  - cfd:", harp_device.syncCfd)
    print ("  - offset:", harp_device.syncChannelOffset)
    print ("  - rate:", harp_device.syncRate)
    print ("  - period:", harp_device.syncPeriod)


def test_orthoflags():
    # checks whether the orthogonal flags class works
    wrn = harp.OrthogonalFlagsMap(0x441, harp.HarpWarnings)
    flg = harp.OrthogonalFlagsMap(0x441, harp.HarpFlags)
    assert len(wrn) == 3
    assert len(flg) == 1
    assert flg.OVERFLOW == True
    assert flg.FIFOFULL == False
    assert isinstance(flg.text('OVERFLOW'), str)
    assert type(flg.mask('OVERFLOW')) == int


def test_dostuff(harp_device):

    # Do stuff.
    harp_device.calibrate()

    # do stuff with channels
    print ("Testing input channel writing...")
    for c in harp_device.channels:
        c.enabled = True
        assert (c.enabled)

        c.offset = 5e-12
        assert (c.offset < 7e-12)

        c.cfd = (0, 0)
        assert len(c.cfd) == 2

        c.enabled = False

    harp_device.syncDiv = 1
    assert harp_device.syncDiv == 1

    harp_device.syncCfd = (1, 0)
    assert harp_device.syncCfd['level'] == 1

    harp_device.syncChannelOffset = 3e-9
    assert harp_device.syncChannelOffset < 5e-9

    harp_device.stopOverflow = 2000
    assert harp_device.stopOverflow == 2000

    harp_device.binning = 1
    assert harp_device.binning == 1

    harp_device.offset = 35e-9
    assert harp_device.offset < 50e-9

    with pytest.raises(harp.HarpError):
        # This is likely to fail because of invalid parameter
        # (histrogramm mode inactive / not properly set up)
        harp_device.histogramLength = 1024
        
    harp_device.clearHistogram()
    print ("Histrogram length:", harp_device.histogramLength)
    assert harp_device.histogramLength is None

    assert type(harp_device.acquiring) == bool

    ## The histogram / acquisition stuff is a bit of a mess, since
    ## for this the device would actually have to be properly set
    ## up and produce data. But since this is just a general s/w unit
    ## testing, we don't know in which state the device, and how it's
    ## behaving, which signals it's attached to etc. So we're just
    ## going to go through some of the motions to make sure the code
    ## doesn't crash, but we don't make any assertions / assumptions
    ## about what it'll result into.
    if harp_device.acquiring:
        print ("Stopping acquisition")
        harp_device.acquiring = False
    else:
        print ("Starting acquisition")
        harp_device.acquiring = True
    print ("Acquiring:", harp_device.acquiring)        

    # setMeasurementcontrolControl() / setMeasControl
    # getHistData etc

    
def test_close():
    ## What happens when closing?
    pass



#!/usr/bin/python3

try:
    from . import harp
except:
    import harp

from softioc import softioc, builder, asyncio_dispatcher
import asyncio

import sys, time

import numpy as np

def init_harp():
    dev = harp.Monster().find("first")

    print ("Initializing Harp device", dev.serial)
    dev.initialize(harp.MeasurementMode.Histogramming,
                   harp.ReferenceClock.Internal)

    dev.calibrate()

    dev.syncDivider = 1
    dev.syncCfd = { "level": 0.300, "zerox": 0.04 }
    dev.syncChannelOffset = 0

    ch = dev.channels[0]
    ch.cfd = { "level": 0.300, "zerox": 0.04 }
    ch.offset = 0
    ch.enable = True

    dev.histogramLength = 3
    dev.binning = 13
    dev.offset = 0

    return dev

def harp_info_header(hdev):
    '''
    Prints a bunch of informative stuff.
    '''
    
    print ("Histogram length:", hdev.histogramLength)
    print ("Base resolution:", hdev.baseResolution)
    print ("Real resolution:", hdev.resolution)
    print ("Histogram delta:", hdev.baseResolution[0]*(2**hdev.binning))
    print ("Histogram reach:", hdev.resolution*hdev.histogramLength)
    print ("Sync rate:", hdev.syncRate)
    print ("Count rates:", [ch.countRate for ch in hdev.channels])
    print ("Warnings:", [ w for w in hdev.warnings] )
    print ("Flags:", [ f for f in hdev.flags] )



    

async def update_iocvars():
    while True:
        #ai.set(ai.get() + 1)
        await asyncio.sleep(1)

        
class HarpSettings(object):
    '''
    Need this to provide some callbacks for variables set via EPICS.
    '''
    def __init__(self, dev):
        self.aqfrequency = 1.0 # in Hz
        self.histogramDelta = dev.resolution
        self.histogramOffset = 0.0
        self.histogramReach = self.histogramDelta * dev.histogramLength

    def setAquisitionFrequency(self, v):
        print ("New acquisition time:", 1.0/v, "seconds")
        self.aqfrequency = v
        
        
if __name__ == "__main__":
    
    dev = init_harp()
    harp_info_header(dev)

    devSettings = HarpSettings(dev)

    dispatcher = asyncio_dispatcher.AsyncioDispatcher()
    builder.SetDeviceName("KMC3XPP_RINGSYNC")

    # some records
    # "input" and "output" are from the client's point of view,
    # i.e. an "output" record is one where the client would send
    # a value to a device.

    # Aquisition time for the Harp device
    aqFrequencySetHz = builder.aOut("aqfrequency_sethz", initial_value=10.0, always_update=True,
                                    on_update=devSettings.setAquisitionFrequency)
    aqFrequency   = builder.aIn("aqfrequency")
    devResolution = builder.aIn("resolution", initial_value=dev.resolution)
    devSyncPeriod = builder.aIn("syncPeriod", initial_value=dev.syncPeriod)
    devSyncRate   = builder.aIn("syncRate",   initial_value=dev.syncRate)
    devWarnings   = builder.stringIn("warnings", initial_value="")
    devFlags      = builder.stringIn("flags", initial_value="")
    
    # Signal / waveform
    axis = np.array([devSettings.histogramOffset + i*devSettings.histogramDelta for i in range(dev.histogramLength)])
    ch0Signal    = builder.Waveform("ch0_signal", length=dev.histogramLength, datatype=float)
    ch0XAxis     = builder.Waveform("ch0_xaxis", length=dev.histogramLength)
    ch0XAxis.set(axis)
    ch0XOffset   = builder.aIn("ch0_xoffset", initial_value=devSettings.histogramOffset)
    ch0XDelta    = builder.aIn("ch0_xdelta", initial_value=devSettings.histogramDelta)
    ch0XReach    = builder.aIn("ch0_xreach", initial_value=devSettings.histogramReach)
    ch0CountRate = builder.aIn("ch0_countrate")
    ch0Peaktime  = builder.aIn("ch0_peaktime")
    

    builder.LoadDatabase()
    softioc.iocInit(dispatcher)

    dispatcher(update_iocvars)

    ch = dev.channels[0]
    t0 = time.time()
    
    while True:
        dev.acquire(1.0/devSettings.aqfrequency)
        while dev.acquiring:
            time.sleep(1.0/devSettings.aqfrequency * 0.01)
        dev.acqstop()
    
        data = ch.histogram
        #data = np.random.rand(*(ch.histogram.shape))
        
        ch0Signal.set(data)
        ch0Peaktime.set(devSettings.histogramOffset+data.argmax()*devSettings.histogramDelta)
        ch0CountRate.set(ch.countRate)
        
        devSyncPeriod.set(dev.syncPeriod)
        devSyncPeriod.set(dev.syncRate)
        devResolution.set(dev.resolution)

        devWarnings.set(" ".join([ w for w in dev.warnings]))
        devFlags.set(" ".join([ f for f in dev.flags]))
    
        tnow = time.time()
        aqFrequency.set(1.0/(tnow-t0))
        t0 = tnow



#!/usr/bin/python3

import miniquant.harp as harp
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import time
import sys

def harp_init():
    '''
    Initialises Harp for a 1.25 MHz device
    '''
    hdev = harp.Monster().find()
    hdev.initialize(harp.MeasurementMode.Histogramming,
                    harp.ReferenceClock.Internal)

    hdev.calibrate()

    hdev.syncDivider = 1
    hdev.syncCfd = { "level": 0.200, "zerox": 0.01 }
    hdev.syncChannelOffset = 0

    ch = hdev.channels[0]
    ch.cfd = { "level": 0.300, "zerox": 0.01 }
    ch.offset = 0
    ch.enable = True

    hdev.histogramLength = 6
    hdev.binning = 0
    hdev.offset = 0

    return hdev


def harp_info_header(hdev):
    '''
    Prints a bunch of informative stuff.
    '''
    
    print ("Histogram length:", hdev.histogramLength)
    print ("Base resolution:", hdev.baseResolution)
    print ("Real resolution:", hdev.resolution)
    print ("Sync rate:", hdev.syncRate)
    print ("Count rates:", [ch.countRate for ch in hdev.channels])
    print ("Warnings:", [ w for w in hdev.warnings] )
    print ("Flags:", [ f for f in hdev.flags] )
    

def mpl_histo_loop(hdev, fig, aqtime=0.1, axis_index=0, line_index=0, channel=0):
    '''
    Loops endlessly reading histogram and displaying it into the
    specified Matplotlib plot, while also serving the GUI event loop.
    '''
    
    print("Starting plotting interface")
    ax = fig.axes[axis_index] if len(fig.axes)>0 \
        else fig.add_subplot(1, 1, 1)
    
    ax.set_xlim(0, hdev.histogramLength)

    line = ax.lines[line_index] if len(ax.lines)>0 \
        else ax.plot([])[0]

    text  = ax.text(10, 0.0, "Jinkies")

    fig.show(False)

    ch = hdev.channels[channel]

    # For auto-scaling
    xlen = 0
    ymax = 0
    t0 = time.time()

    print("Entering histogram loop, %.5f seconds acquisition time" % aqtime)
    while True:

        hdev.acquire(aqtime)
        while hdev.acquiring:
            fig.canvas.flush_events()
        hdev.acqstop()
    
        data = ch.histogram
    
        if len(data) != xlen:
            xlen = len(data)
            x = np.array(range(xlen))

        ## Some random noise, for testing
        #data[np.random.randint(len(x))] = np.random.randint(100)        

        dmax = data.max()
        if ymax < dmax:
            print ("Adjusting Y range")
            ymax = dmax
            ax.set_ylim(-1, dmax+1)

        tnow = time.time()
        text.set_text("aq: %2.0f Hz | flags: %r | counts: %d" % \
                      (1.0/(tnow-t0),
                       [f for f in hdev.flags],
                       data.sum()))
        t0 = tnow
    
        line.set_data(x, data)
        fig.canvas.draw_idle()
        fig.canvas.flush_events()

    
if __name__ == "__main__":

    print("Initializing device")
    hdev = harp_init()

    time.sleep(0.5)
    harp_info_header(hdev)

    fig = plt.figure()
    
    mpl_histo_loop(hdev, fig)



#!/usr/bin/python3

#
#        MiniQuant --  Wrapper for PicoQuant's HydraHarp access libraries.
#        Copyright (C) 2022 Florin Boariu.
#
#        This program is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        This program is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

#
# Unit-testing the Harp classes with a quirked or unplugged USB
# (i.e. mainly testing for correct error handling / reporting).
#

import pytest

from miniquant import harp

def test_openfail():
    # Fails opening with specific errors, depending on
    # whether unplugged or quirked.

    mon = harp.Monster()
    with pytest.raises(OSError) as e:
        h3 = mon.find()



#!/usr/bin/python3

#
#        MiniQuant --  Wrapper for PicoQuant's HydraHarp access libraries.
#        Copyright (C) 2022 Florin Boariu.
#
#        This program is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        This program is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import ctypes
from ctypes import byref
from os import environ as env
from os import strerror
from time import sleep
import errno
from numpy.ctypeslib import as_array
from numpy import zeros

from threading import Lock

import logging

class HarpError(Exception):
    '''
    Base class for reporting errors of a Harp device
    '''
    def __init__(self, harpcode=-1, message=None):
        self.harpcode = harpcode
        self.message = message

    def __str__(self):
        return "Harp error: %s (%d)" % (self.message or "*arrrgh!*", self.harpcode)

    def __repr__(self):
        return str(self)

    
class ErrorStringPortal:
    '''
    Wrapper class for GetErrorString(). The purpose is to provide
    an indexing interface which can mimic error strings as being
    an array or a dictionary.
    '''
      
    def __init__(self, hlib):
        self.hlib = hlib
            
    def __getitem__(self, err):
        buf = ctypes.create_string_buffer(b'', 256)
        self.hlib.HH_GetErrorString(buf, err)
        return buf.value.decode('utf-8')

    def __call__(self, ret):
        '''
        Checks `ret` for success value (`==0`) or raises an exception
        containing the error message.
        '''
        if ret < 0:
            raise HarpError(harpcode=ret, message=self[ret] if self.hlib else None)
        
        
class MeasurementControl:
    '''
    Constants / settings for HH_SetM2022-12-10 13:03:24,492 INFO: histogram: has become available again                                                                                             
easControl. See
    [HydraHarp 400 Documentation](https://raw.githubusercontent.com/PicoQuant/HH400-v3.x-Demos/master/Linux/manual_HHLibLinuxv3.0.0.2.pdf).
    '''
    ContC1Gated = 4
    ContC1StartCtcStop = 5
    ContCtcRestart = 6

class MeasurementMode:
    '''
    Constants for HH_Initialize `mode` parameter
    '''
    Histogramming = 0
    T2 = 2
    T3 = 3
    Continuous = 8


class MeasurementControl:
    '''
    Constants for `HH_SetMeasControl`.
    '''
    SingleShotCtc = 0
    C1Gated = 1
    C1StartCtcStop = 2
    C1StartC2Stop = 3
    ContC1Gated = 4
    ContC1StartCtcStop = 5
    ContCtcRestart = 6


class EdgeSelection:
    '''
    Constants for ede determination, used in `HH_SetMeasControl`.
    '''
    Falling = 0
    Rising = 1


class ReferenceClock:
    '''
    Constants for HH_Initialize `refsource`.
    '''
    Internal = 0
    External = 1

class DeviceLimits:
    '''
    Some hard-coded maximal legths and other devicel limits.
    May or may not change...?
    '''
    MaxHistogramLength = 65536    ## MAXHISTLEN
    MaxHistogramKBase  = 6        ## MAXLENCODE (the base for exponent 2 as a KB histo length)
    MaxDevices         = 8        ## MAXDEVNUM
    MaxBinning         = 26       ## MAXBINSTEPS
    CModeMaxHistogramLength = 8192   ## MAXHISTLEN_CONT
    CModeMaxHistogramKBase  = 3      ## MAXLENCODE_CONT
    CModeMaxBufferLength    = 262272 ## MAXCONTMODEBUFLEN
    MaxFifoReadout     = 131072   ## TTREADMAX
    MinFifoReadout     = 128      ## TTREADMIN
    MaxBytesReadout   = 262272    ## MAXCONTMODEBUFLEN
#
# Warning texts from the HHLIB API definition.
#
# This is a map of warning masks to strings.
# We save two strings: the "official" API constant name for recognition,
# and the unofficial "plain English" translation. Most of the time they're
# the same, but they may not be.
HarpWarnings = {
    'SYNC_RATE_ZERO':      (0x001, 'Sync rate is zero'),
    'SYNC_RATE_TOO_LOW':   (0x002, 'Sync rate is too low'),
    'SYNC_RATE_TOO_HIGH':  (0x004, 'Sync rate is too high'),
    'INPT_RATE_ZERO':      (0x010, 'Input rate is zero'),
    'INPT_RATE_TOO_HIGH':  (0x040, 'Input rate is too high'),
    'INPT_RATE_RATIO':     (0x100, 'Bad input rate ratio'),
    'DIVIDER_GREATER_ONE': (0x200, 'Rate divider is greater than 1'),
    'TIME_SPAN_TOO_SMALL': (0x400, 'Time span is too small'),
    'OFFSET_UNNECESSARY':  (0x800, 'Offset unnecessary') }


#
# Flags as specified in the HHLIB API for current status
#
HarpFlags = {
    'OVERFLOW':  (0x0001, "Aquisiton overflow"),
    'FIFOFULL':  (0x0002, "FIFO buffer is full"),
    'SYNC_LOST': (0x0004, "Lost synchronisation"),
    'REF_LOST':  (0x0008, "Lost reference signal"),
    'SYSERROR':  (0x0010, "You broke it, you pay for it"),
    'ACTIVE':    (0x0020, "Data aquisition in process") }

#
# Features as specified in the HHLIB API (a.k.a. device capabilities)
#
HarpFeatures = {
    'DLL':     (0x01, "DLL Mode"),
    'TTTR':    (0x02, "TTTR Mode"),
    'MARKERS': (0x04, "Markers"),
    'LOWRES':  (0x08, "Low resolution mode"),
    'TRIGOUT': (0x10, "Trigger output") }


class OrthogonalFlagsMap:
    '''
    This is a mapping helper for various flags-to-strings facilities
    (Harp warnings, Harp flags, ...) encoded as a i.e. a bitwise field
    of *several* items. You can iterate through the warnings to reach
    all of them:
    ```
       >>> warn = OrthogonalFlagsMap(HarpWarnings, 0x441)
       >>> [ warn.text(w) for w in warn ]
       [ 'Sync rate is zero', 'Input rate is too high', 'Time span is too small' ]
       >>> warn.INPT_RATE_ZERO
       True
       >>> warn.INPT_RATE_TOO_HIGH
       False
    ```
    '''
    
    def __init__(self, flagsMap, code=None):
        '''
        Initializes the mapper. Parameters:
          - `code`: This is a bitwise field of flags that this instance represents.
          - `flagsMap`: This is a dictionary which maps single string keys to
            `(flagMask, flagDescription)` tuples. The string key part
            is a non-changeable string that describes the flags for all eternity,
            and which the user (or other program layers) can use to access the flag.
            `flagDescription` is a human-readable string which can change, e.g.
            by translation or a more precise specification, and `flagMask` is a bit
            mask that indicates whether the flag is set or not.
        '''
        self.flagsMap = flagsMap
        if code:
            self.recode(code)
    
    def recode(self, code):
        '''
        Resets the code to `code` and returns a reference to self.
        This is to update the active/inactive flags list to the
        ones encoded in `code` without altering the identity of the
        object.
        '''
        self.code = code
        return self

    def __str__(self):
        return str([f for f in self.keys()])

    def __getattr__(self, key):
        return (self.code & self.flagsMap[key][0]) != 0

    def __iter__(self):
        ''' Iterate through all warnings encoded in `self.code`. '''
        for k,v in self.flagsMap.items():
            if (v[0] & self.code) != 0:
                yield k

    def keys(self):
        '''
        Mimic a bit of a `dict`-like interface: return all the HHLIB API
        warning keys that are encoded in `self.code`.
        '''
        for k in self:
            yield k

    def items(self):
        '''
        Mimic a bit more a `dict`-like interface: return all the HHLIB API
        warning keys that are encoded in `self.code`.
        '''
        for k,v in self.flagsMap.items():
            if (v[0] & self.code):
                yield (k, v[1])
    
    def __getitem__(self, flag):
        ''' Another way of reading a flag '''                                                 
        return self.__getattr__(flag)

    def text(self, flag):
        '''
        Returns the description text.
        '''
        return self.flagsMap.get(flag, None)[1]

    def mask(self, flag):
        '''
        Returns the numerical mask value.
        '''
        return self.flagsMap.get(flag, None)[0]

    def __len__(self):
        return len([i for i in self.items()])
    

class ModuleInfoPortal:
    '''
    Wrapper class for `HH_GetModuleInfo`. The purpose is to provide
    an indexing interface which can mimic module info as being
    an array.
    '''

    class ModuleInfo:
        '''
        The class with the actuar getter code for HH_GetModuleInfo
        '''
        def __init__(self, harp, modindex):
            self.harp = harp
            self.mod = modindex

        @property
        def info(self):
            code = ctypes.c_int()
            ver  = ctypes.c_int()
            self.harp.hlib.HH_GetModuleInfo(self.harp.devid, self.mod, byref(code), byref(ver))
            return { "code": code.value, "version": ver.value }

            meascontrol
        
    def __init__(self, harp):
        self.harp = harp

    def __iter__(self, *args):
        l = len(self)
        for i in range(l):
            yield self[i]            
            
    def __getitem__(self, mod):
        return self.ModuleInfo(self.harp, mod)
        
    def __len__(self):
        n = ctypes.c_int()
        self.harp.hlib.HH_GetNumOfModules(self.harp.devid, ctypes.byref(n))
        return n.value


class InputChannelPortal:
    '''
    Wrapper class for various channel-based classes The purpose is to provide
    an indexing interface which can mimic module info as being an array.
    Each element in that array is a Channel and has attributes that correspond
    to getter and setter functions like:
      - HH_GetModuleIndex
      - HH_SetInputCFD
      - HH_SetInputChannelOffset
      - HH_SetInputChannelEnable
      - HH_GetHistogram
      - HH_GetCountRate
    '''

    class InputChannel:
        '''
        The actual Channel class that has all the getters/setters.
        '''
        def __init__(self, harp, channel):

            self.harp = harp
            #self.hlib = harp.hlib           # HHLIB handler to talk to
            #self.devid = harp.devid         # Harp device we're using
            self.channel = channel          # that's us, the channel ID
            self.clearOnReadout = False

        def __str__(self):
            return "%r:ch[%d]" % (self.harp, self.channel)

        
        def __repr__(self):
            return str(self)
            

        @property
        def moduleIndex(self):
            ''' module which serves this channel '''
            modofc = ctypes.c_int()
            self.harp.hlib_request('HH_GetModuleIndex', self.channel, byref(modofc))
            return modofc.value

        @property
        def enabled(self):
            return self.harp.shadow('ch%d-enabled' % self.channel, None)

        @enabled.setter
        def enabled(self, val):
            '''
            Wrapper for `HH_SetInputChannelEnable`. Working with a shadow value
            for artificial getter.
            '''
            enbl = 1 if val else 0
            self.harp.hlib_request('HH_SetInputChannelEnable', self.channel, enbl,
                                   shadow='ch%d-enabled' % self.channel, value=bool(val))


        @property
        def offset(self):
            ''' returns last value from shadow cache '''
            return self.harp.shadow('ch%d-offset' % self.channel, None)

        @offset.setter
        def offset(self, value):
            '''
            Wrapper for `HH_SetInputChannelOffset`.
            Working with a shadow value for artificial getter.
            Time offset value here is seconds (floating point),
            opposed to picoseconds of `HH_SetInputChannelOffset`.
            '''
            self.harp.hlib_request('HH_SetInputChannelOffset', self.channel, int(value*1e12),
                                   shadow='ch%d-offset' % self.channel, value=value)

            
        @property
        def cfd(self):
            ''' returns CFD from shadow '''
            return self.harp.shadow('ch%d-cfd' % self.channel, None)

        @cfd.setter
        def cfd(self, data):
            '''
            Sets the CFD -- data is either a tuple `(level, zerox)` or a dict
            `{"value": ..., "zerox": ...}`. This is a wrapper for `HH_SetInputCFD`.
            '''
            if isinstance(data, dict):
                d = data
            else:
                d = dict({"level": data[0], "zerox": data[1]})
            ret = self.harp.hlib_request('HH_SetInputCFD', self.channel,
                                         int(d["level"]*1e3), int(d["zerox"]*1e3),
                                         shadow='ch%d-cfd' % self.channel, value = d)


        @property
        def countRate(self):
            ''' Wrapper for `HH_GetCountRate`. '''
            s = ctypes.c_int()
            self.harp.hlib_request('HH_GetCountRate', self.channel, byref(s))
            return s.value


        @property
        def histogram(self):
            '''
            Returns an ndarray with the current histogram data for this channel.
            Clears the acquisition buffer after data retrieval.
            '''
            numpts = self.harp.feedback.get('histogramLength',
                                             DeviceLimits.MaxHistogramLength) 

            #logging.debug("Flags: %r" % ([f for f in self.harp.flags],))
            
            #if self.harp.acquiring:
            #	return None #zeros([numpts])
            #numpts_param = ctypes.c_int(numpts)
            
            erase = ctypes.c_int32(int(self.clearOnReadout))
            dbuf = (ctypes.c_int32*numpts)()
            self.harp.hlib_request('HH_GetHistogram', dbuf, self.channel, erase)
            data = as_array(dbuf)
            logging.debug("Histogram readout with %d counts total in %d channels, still acquiring: %r" \
            	% (data.sum(), data.shape[0], self.harp.acquiring))
            return data
            
        
    def __init__(self, harp):
        self.harp = harp

    def __iter__(self, *args):
        l = len(self)
        for i in range(l):
            yield self[i]            
            
    def __getitem__(self, ch):
        return self.InputChannel(self.harp, ch)

    def __len__(self):
        n = ctypes.c_int()
        self.harp.hlib_request('HH_GetNumOfInputChannels', ctypes.byref(n))
        return n.value
    

class MonsterHarp(object):
    '''
    Manages one specific HydraHarp device (i.e. one of the `HH_OpenDevice()`
    thingies that has an integer index).
    Offers all the `HH_...` functions of the HydraHarp library with a more
    pythonic interface.
    '''

    class ReferenceCounter:
        ''' Used to know when to close the device '''
        def __init__(self):
            self.counter = 1

        def inc(self):
            self.counter += 1
            return self.counter

        def dec(self):
            self.counter -= 1
            return self.counter

        def __bool__(self):
            return (self.counter > 0)


    def __init__(self, id=0, hlib=None, mt=True):
        '''
        Initializes / opens the device with the specified id.
        
        Parameters:
        
          - `id`: This is an integer to represent which device to open,
            from the point of view of the HH-library. It is unclear how
            the library tells different devices apart and which one
            has which ID; however, apparently valid IDs go from 0 to 7,
            and usually only one is available when only one USB device
            is attached.
        
          - `hlib`: This is a `Monster` object, which is a wrapper around
            a `ctypes.CDLL()` object around the HH-library.
            The `Monster` class is a singleton, i.e. there is only one
            instance available in the system.
            See the `Monster` documentation about how the first default
            instance is initialized.
            You can grab one by calling its constructor (`m = Monster()`).

            Generally, `MonsterHarp` doesn't need an explicit `hlib` parameter
            set, the default being the module's `monster_binding` object.
            But one *can* be specified for... ahem... advanced needs :-)
                
        '''

        # If this is set, access to hlib will be protected my a mutex.
        self._lock = Lock() if mt else None
        
        self._hlib = hlib or Monster().hhlib ##monster_binding.hhlib

        # This is for the __copy__ / __deepcopy__ operator.
        # We need to reference everything from the source object.
        if id < 0:
            return
        
        sbuf = ctypes.create_string_buffer(b'', 16)
        self.refcnt = self.ReferenceCounter()
        ret = self.hlib.HH_OpenDevice(id, sbuf)

        self._shadow = {}
        self.feedback = {}

        if ret == 0:
            self.serial = sbuf.value.decode('utf-8')
            self.devid = id
            
        elif ret == -1:
            errcode = ctypes.get_errno()
            errstr = strerror(errcode)
            if errcode == errno.ETIMEDOUT:
                errstr += " (crappy hardware 0e0d:0009:k?)"
                self.devid = -1
            raise OSError(errcode, errstr)

        else:
            raise RuntimeError("This isn't supposed to return %d" % ret)

    @property
    def clearOnAcquire(self):
        '''
        Automatically clears histogram when acquisition is started
        '''
        return self._shadow.setdefault('clearOnAcquire', True)

    @clearOnAcquire.setter
    def clearOnAcquire(self, val):
        self._shadow['clearOnAcquire'] = val

    @property
    def hlib(self):
        '''
        Returns a reference to the HLIB object we're using, optionally protecting
        calls with a mutex (if the object was initianized with mt=True).
        '''
        return self._hlib

    def hlib_request(self, req, *params, shadow=None, value=None):
        '''
        Executes a lock-protected HLIB request and returns the result.
        If `shadow` is not None, the shadow key `shadow` will be set to
        value `value` while still undmeascontroler mutex protection.
        '''
        requestProc = getattr(self.hlib, req)
        logging.debug("HLIB request: %r" % (req))
        if self._lock:
            with self._lock:
                ret = requestProc(self.devid, *params)
                ErrorStringPortal(self.hlib)(ret)
                if shadow is not None:
                    self._shadow[shadow] = value
                return ret
        else:
            ret = requestProc(self.devid, *params)
            ErrorStringPortal(self.hlib)(ret)
            if shadow is not None:
                self._shadow[shadow] = value            
            return ret
    
        
    def __del__(self):
        if self.refcnt.dec():
            return
        
        if hasattr(self, 'devid') and self.devid >= 0:
            self.hlib.HH_CloseDevice(self.devid)
            self.refcnt = None
            self.devid  = None
            self.serial = None
            self._hlib   = None
            self._shadow = None
            self.feedback = None

            
    def __copy__(self):
        # Upon copy/deepcopy we take over references of the ID and refcounter
        # of the source object. Once opened, we don't want to prematurely
        # close the USB device (hhlib doesn't account for these types of
        # concurrencies.)
        copy = MonsterHarp(hlib)
        copy.serial = self.serial
        copy.devid  = self.devid
        copy.refcnt = self.refcnt
        copy._shadow = self._shadow
        copy.feedback = self.feedback
        copy.refcnt.inc()
        return copy

    
    def __deepcopy__(self):
        return self.__copy__()


    def __str__(self):
        return "HydraHarp[%s]" % (self.serial)


    def __repr__(self):
        return str(self)
    

    def _checkHarpError(self, ret):
        '''
        To simplify error handling: raises an exception if ret < -1.
        '''
        ErrorStringPortal(self.hlib)(ret)
        

    def initialize(self, mode, clock):
        '''
        Wrapper for `HH_Initialize`. Call before any other routines. Parameters:
        - `mode`: See `MeasurementMode`
        - `clock`: `ReferenceClock.Internal` or `ReferenceClock.External`
        Returns `None` on success, or raises 
        '''
        self.hlib.HH_Initialize(self.devid, mode, clock)


    @property
    def hardwareInfo(self):
        '''
        Wrapper for `HH_GetHardwareInfo`. Returns a string tuple of model, partno, version.
        '''
        mod = ctypes.create_string_buffer(b'', 64)
        pno = ctypes.create_string_buffer(b'', 64)
        ver = ctypes.create_string_buffer(b'', 64)
        self.hlib_request('HH_GetHardwareInfo', mod, pno, ver)
        return { "model": mod.value.decode('utf-8'),
                 "part": pno.value.decode('utf-8'),
                 "vesion": ver.value.decode('utf-8') }
    
    
    @property
    def serialNumber(self):
        '''
        Reads out the serial number. Should return the same as `self.serial`, but
        the difference between `self.serial` and this is that `serialNumber` actually
        does query the serial number via the HHLIB API on every call, while `self.serial`
        is the number that was cached when the device was first opened.
        '''
        s = ctypes.create_string_buffer(b'', 64)
        self.hlib_request('HH_GetSerialNumber', s)
        return s.value.decode('utf-8')

    
    @property
    def features(self):
        '''
        Wrapper for `HH_GetFeatures`. Returns a 32-bit bitmask with features.
        See PicoQuant documentation for the meaning of the bits.
        '''
        i = ctypes.c_int()
        self.hlib_request('HH_GetFeatures', ctypes.byref(i))
        m = self._shadow.setdefault('featuresMap', OrthogonalFlagsMap(HarpFeatures))
        m.recode(code=i.value)
        return m

    
    @property
    def baseResolution(self):
        '''
        Wrapper for `HH_GetBaseResolution`. Returns a tuple consisting of
        a floating-point number with resolution in seconds (ACHTUNG, this is
        different from the origina HHLIB API!) and the maximum allowed number
        of binning steps.
        '''

        d = ctypes.c_double()
        i = ctypes.c_int()
        self.hlib_request('HH_GetBaseResolution', ctypes.byref(d), ctypes.byref(i))
        return d.value*1e-12, i.value
    

    @property
    def channels(self):
        ''' Portal for all input channel related attributes '''
        return InputChannelPortal(self)

    
    @property
    def modules(self):
        '''
        (Preparation for) wrapper for `HH_GetModuleInfo`. Returns
        an object that can iterate through module information.
        '''
        return ModuleInfoPortal(self)

    
    @property
    def hardwareDebugInfo(self):
        '''
        Wrapper around `HH_HardwareDebugInfo`. Apparently can be queried if
        HH_ERROR_STATUS_FAIL or FLAG_SYSERROR is encountered.
        Returns a python utf-8 string.
        '''
        s = ctypes.create_string_buffer(b'', 65636)
        self.hlib_request('HH_GetHardwareDebugInfo', s)
        return s.value.decode('utf-8')

    
    def calibrate(self):
        '''
        Wrapper around `HH_Calibrate`. Does something. No return value.
        '''
        self.hlib_request('HH_Calibrate')


    def shadow(self, key, default=None):
        if self._lock:
            with self._lock:
                return self._shadow.get(key) or default
        else:
            return self._shadow.get(key) or default
        

    @property
    def syncDiv(self):
        return self.shadow('syncDiv', None)

    @syncDiv.setter
    def syncDiv(self, val):
        '''
        Wrapper for `HH_SetSyncDiv`. Working with a shadow value for artificial getter.
        '''
        p = ctypes.c_int(val)
        self.hlib_request('HH_SetSyncDiv', p, shadow='syncDiv', value=val)

        
    @property
    def syncCfd(self):
        return self.shadow('syncCfd', None)

    @syncCfd.setter
    def syncCfd(self, data):
        '''
        Wrapper for `HH_SetSyncCfd`. Working with a shadow value for artificial getter.
        `data` is either a tuple `(level, zerox)` or a dict `{"level": ..., "zerox": ...}`.
        ACHTUNG, voltage levels are in V, not mV.
        '''
        if isinstance(data, dict):
            d = data
        else:
            d = dict({"level": data[0], "zerox": data[1]})
        p0 = ctypes.c_int(int(d["level"]*1e3))
        p1 = ctypes.c_int(int(d["zerox"]*1e3))
        self.hlib_request('HH_SetSyncCFD', p0, p1, shadow='syncCfd', value=d)


    @property
    def syncChannelOffset(self):
        return self.shadow('syncChannelOffset', None)

    @syncChannelOffset.setter
    def syncChannelOffset(self, value):
        '''
        Wrapper for `HH_SetSyncChannelOffset`.
        Working with a shadow value for artificial getter.
        '''
        p = ctypes.c_int(int(value*1e12))
        self.hlib_request('HH_SetSyncChannelOffset', p,
                          shadow='syncChannelOffset', value=value)

    @property
    def stopOverflow(self):
        '''
        Returns the stop overflow value from shadow. The value is an integer
        if stop overflow is enabled, `False` if disabled, and `None` if no
        value has yet been set (in which case the device is in an unknown
        state).
        '''
        return self.shadow('stopOverflow', None)

    @stopOverflow.setter
    def stopOverflow(self, value):
        '''
        Wrapper for `HH_SetStopOverflow`. Unlike its C counterpart, this only
        takes one argument: `None` or `False` for disabling stop overflow
        (in which case `HH_SetStopOverflow` will be called with a 0 for its
        `stop_ofl` parameter), or an unsigned integer count at which measurement
        should be stopped in case of overflow.
        '''
        stop_ofl = 0 if not bool(value) else 1
        stop_cnt = 0 if stop_ofl==0 else value
        ret = self.hlib_request('HH_SetStopOverflow', stop_ofl, stop_cnt,
                                shadow='stopOverflow',
                                value=(value if stop_ofl == 1 else False))


    @property
    def binning(self):
        ''' returns last binning value from shadow '''
        return self.shadow('binning', None)

    @binning.setter
    def binning(self, value):
        ''' Wrapper for `HH_SetBinning`, value is base 2. '''
        ret = self.hlib_request('HH_SetBinning', value,
                                shadow='binning', value=value)

    
    @property
    def offset(self):
        ''' returns last offset value from shadow (ACHTUNG, NOT inputChannelOffset!)'''
        return self.shadow('offset', None)
    
    @offset.setter
    def offset(self, value):
        '''
        Wrapper for `HH_SetOffset`. Offset time here is specified
        as floating point in seconds (unlike `HH_SetOffset`
        which is nanoseconds).
        '''
        p = ctypes.c_int(int(value*1e9))
        ret = self.hlib_request('HH_SetOffset', p,
                                shadow='offset', value=value)
        return value


    @property
    def histogramLength(self):
        '''
        Returns histrogram length from feedback cache. ACHTUNG, this
        is not the shadow cache (it acts similarly), it's what the
        device actually transmitted on the last setting.
        '''
        return self.feedback.get('histogramLength', None)

    @histogramLength.setter
    def histogramLength(self, base):
        ''' Wrapper for `HH_SetHistoLen`. '''
        actual = ctypes.c_int()
        ret = self.hlib_request('HH_SetHistoLen', base, byref(actual))
        self.feedback['histogramLength'] = actual.value
        return actual.value

        
    def clearHistogram(self):
        '''
        Wrapper for `HH_ClearHistMem`.
        '''
        ret = self.hlib_request('HH_ClearHistMem')

    @property
    def measurementControl(self):
        ''' returns the measurement control form shadow, or `None` if none was set yet '''
        return self.shadow('measurementControl', None)

    @measurementControl.setter
    def measurementControl(self, data):
        '''meascontrol
        Wrapper for `HH_SetMeasControl`. Working with a shadow value for artificial getter.
        `data` is either a tuple `(ctrl, startedge, stopedge)` or a dict
        `{"control": ..., "startEdge": ..., "stopEdge": ...}`.
        '''
        if isinstance(data, dict):        
           idx = ["control", "startEdge", "stopEdge"]
        else:
           idx = (0, 1, 2)
        
        mc_ctrl = getattr(MeasurementControl, data[idx[0]]) \
                if isinstance(data["control"], str) else data[idx[0]]
                
        [mc_start, mc_stop] = [ getattr(EdgeSelection, data[i]) \
              if isinstance(data[i], str) else data[i] for i in idx[1:] ]
	    
        #print(mc_ctrl, mc_start, mc_stop)
	    
        d = { "control":   mc_ctrl,
              "startEdge": mc_start, 
              "stopEdge":  mc_stop}
              
        logging.debug("Measurement control: %r" % d)

        ret = self.hlib_request('HH_SetMeasControl', mc_ctrl, mc_start, mc_stop,
                                shadow='measurementControl', value=d)

        
    def acquire(self, aqtime):
        '''
        Wrapper for `HH_StartMeas`. ACHTUNG, aquisition time here is in seconds(!).
        '''
        self.aquisitionTime = aqtime
        if self.clearOnAcquire:
            logging.debug("Clearing histogram")
            self.clearHistogram()
        logging.debug("Measurement control: %r" % self.measurementControl)
        v = ctypes.c_int(int(aqtime*1e3))
        self.hlib_request('HH_StartMeas', v)

        
    def acqstop(self):
        '''
        Wrapper for `HH_StopMeas`.
        '''
        self.hlib_request('HH_StopMeas')
        

    @property
    def acquisitionTime(self):
        '''
        Returns the last aquisition time that was set, or a default of 1.0 seconds
        '''
        return self._shadow.setdefault('acquisitionTime', 1.0)


    @acquisitionTime.setter
    def acquisitionTime(self, val):
        self._shadow['acquisitionTime'] = val


    @property
    def acquiring(self):
        '''
        Wrapper for `HH_CTCStatus`. Returns `True` if the aquisition is still running,
        `False` otherwise. Setting this to `False` also stops the current measurement.
        '''
        s = ctypes.c_int()
        self.hlib_request('HH_CTCStatus', byref(s))
        return (s.value == 0)

    @acquiring.setter
    def acquiring(self, enable):
        '''
        If this is set to True, it triggers an acuisition with the last
        acquisition time, or with the default of 1 second. If this is set
        to False, it cancels the current aquisition.
        '''
        if enable:
            if self.acquiring:
                #logging.warning("Acquisition already in progress, please stop first")
                return
            t = self.acquisitionTime
            logging.debug("Starting aquisition with %r seconds (enable: %r)" % (t, enable))
            self.acquire(t)
        else:
            logging.debug("Stopping aquisition (requested: %r, running: %r)" \
                          % (enable, self.acquiring))
            self.acqstop()


    @property
    def resolution(self):
        ''' Wrapper for `HH_GetResolution`. Returns the time resolution in seconds (!) '''
        s = ctypes.c_double(-1.0)
        self.hlib_request('HH_GetResolution', byref(s))
        return (s.value * 1e-12)


    @property
    def syncRate(self):
        ''' Wrapper for `HH_GetSyncRate`. '''
        s = ctypes.c_int()
        self.hlib_request('HH_GetSyncRate', byref(s))
        return s.value


    @property
    def syncPeriod(self):
        ''' Wrapper for `HH_GetSyncPeriod`. Returns the sync period in seconds (!). '''
        s = ctypes.c_float()
        self.hlib_request('HH_GetSyncPeriod', byref(s))
        return s.value * 1e-12


    @property
    def warnings(self):
        '''
        Returns the warnings. This is mainly a wrapper for `HH_GetWarnings`, but HHLIB
        documentation states that on all channels `HH_GetCountRate` needs to be called
        for all channels beforehand. Probably also `HH_GetSyncRate`.
        '''
        s = ctypes.c_int()
        cnt = [ c.countRate for c in self.channels ]
        sr = self.syncRate
        self.hlib_request('HH_GetWarnings', byref(s))
        m = self._shadow.setdefault('warningsMap', OrthogonalFlagsMap(HarpWarnings))
        m.recode(s.value)
        return m


    @property
    def flags(self):
        '''
        Returns the flags. This is a wrapper for `HH_GetFlags`.
        '''
        s = ctypes.c_int()
        self.hlib_request('HH_GetFlags', byref(s))
        m = self._shadow.setdefault('flagsMap', OrthogonalFlagsMap(HarpFlags))
        m.recode(s.value)
        return m


    @property
    def markers(self):
        '''
        Returns the markers setup. This is a list of size 4 which holds
        either an `EdgeSelection` object (Rising/Falling), or `None`,
        one for each of the 4 possible markers that a Harp device manages.
        If the entry is `None`, the edge is deactivated.
        The getter returns the last setting from shadow memory, the
        setter is a wrapper for `HH_SetMarkerEdges` / `HH_SetMarkerEnable`.
        If the markers were never set, then the getter returns only `None`
        instead of a list of 4 items (the shadow memory being empty).
        '''
        return self.shadow('markers', None)

    @markers.setter
    def markers(self, status):
        '''
        Sets the edge specification and enables/disables markers.
        The parameter `status` is an enumerable (list, tuple, array)
        with exactly 4 items. Each of the items must be either
        an `EdgeSelection` object, or the keyword `None`.
        If the item is `None`, the marker is deactivated.
        '''
        edges = [ ctypes.c_int(1 if s==EdgeSelection.Rising else 0) for s in status ]
        enablers = [ ctypes.c_int(1 if s is not None else 0) for s in status ]
        self.hlib_request('HH_SetMarkerEdges', *edges),
        self.hlib_request('HH_SetMarkerEnable', *enablers)


    @property
    def markerHoldoffTime(self):
        ''' returns the last markerHoldoffTime from shadow '''
        return self.shadow('markerHoldoffTime', None)
        
    @markerHoldoffTime.setter
    def markerHoldoffTime(self, t):
        '''
        Wrapper for `HH_SetMarkerHoldoffTime`. `t` is time in nanoseconds (integer).
        '''
        s = ctypes.c_int(t)
        self.hlib_request('HH_SetMarkerHoldoffTime', s)


    @property
    def fifo(self):
        '''
        Returns the FIFO contents as much as they are available as
        a numpy array (typically it's a multiple of 128 items).
        This is a wrapper for `HH_ReadFiFo`.
        '''
        nummax = ctypes.c_int(DeviceLimits.MaxFifoReadout)
        numread = ctypes.c_int()
        rbuf = (ctypes.c_int*nummax)()
        self.hlib_request('HH_ReadFiFo', rbuf, nummax, byref(numread))
        return as_array(rbuf)[:numread.value]
        

class Monster(object):
    '''
    Wrapper around PicoQuant's HydraHarp C library, typically called hhlib.so
    on Linux systems.

    This package does *not* distribute any code, Imaginary or Inelectual
    Property of PicoQuant GmbH -- for that, check their GitHub pages
    (specifically that of the
    [HydraHarp 400](https://github.com/PicoQuant/HH400-v3.x-Demos)). You
    will need to download the binaries yourself for this to work,
    and set the `MINIQUANT_HHLIB` enviromnet variable to the
    name, preferrably including full path, of PicoQuant's proprietary
    binary.

    This is a Singleton, by default initialized on `import ...` using the
    ``harp.monster_library` shared object path (which in turn is
    extracted from `MINIQUANT_HHLIB`). If the default initialization
    fails, it does so silently. The most likely reason is that the
    correct HH-library wasn't found or couldn't be loaded.
    Then the user has to initialize the class at least once
    correctly and explicitly pass that instance to any `MonsterHarp`
    initializations, e.g.:
    ```
    # library loading triggers Monster initialisation from env-var
    from harp import Monster, MonsterHarp

    # scenario 1a: implicit intialisation worked; open a harp now
    # by passing an explicit ID to Monsterharp
    h = MonsterHarp(0)

    # scenario 1b: implicit initialization worked; find a harp
    # by obtaining a copy of the Monster singleton first
    mcopy = Monster()
    h = mcopy.find("any")

    # scenario 2a: implicit initialisation failed; re-initialise
    # the HH-lib implicitly and open a harp by creating a Harp object
    m = Monster(library="/path/to/real/hhlib.so")
    h = MonsterHarp(0, hlib=m.hlib)

    # scenario 2b: ...or find a harp by using the singleton
    h = m.find("any")
    ```

    In each of the examples above there is always only one `Monster`
    instance alive.
    '''

    def __new__(self, *args, **kwargs):
        # Make a singleton
        if not hasattr(self, 'instance'):
            self.instance = super(Monster, self).__new__(self)            
        return self.instance
    

    def __init__(self, library=None):
        '''
        Initializes the library / Harp access system. This is just
        a precursor to using the Harp.

        To actually start connecting to one of the devices, see
        `connect()`.
        '''

        lib = env.get('MINIQUANT_HHLIB', "hhlib.so")
        if not hasattr(self, 'hlib'):
            try:
               l = library or lib
               self.hlib = ctypes.CDLL(l, use_errno=True)
            except OSError:
                logging.error("Error opening: %r" % l)
                raise


    def find(self, *serials, maxdev=8):
        '''
        Returns a `MonsterHarp` object (or a list of objects)
        which can be used to talk to one specific PicoQuant
        HydraHarp device.

        Parameters:
          - `serials`: This a list of strings,
             or one of the special values `None`, `all`, or `first`.
        
            If this is a string, the HydraHarp device with
            the given serial number will be opened and a corresponding
            `MonsterHarp` object is returned.

            If this is a list of strings, then a map will be returned --
            one for each device that was found and that has a serial
            number that matches any of the strings in the list.

            If the value is `None` or `"first"`, then the first device
            that will be found is returned.

            If it is `"all"`, then a dictionary with all devices
            (serials as keys) is returned.

            The device serial numbers are all supposed to be strings.

          - `maxdev`: The maximum number of devices to try opening.
            8 is a hardcoded value in the PicoQuant's library. There's
            no harm in trying higher numbers, but it most likely won't
            have any effect.

        Returns:
          - a single `MonsterHarp` object if a single serial number,
            or `serial="any"` was requested and found.
          - a dictionary of `MonsterHarp` objects if more than one
            serial number was requested and found, or `serial="all"`
            was requested.
          - an empty list if no device was specifically requested (e.g.
            the call was `serial="all"`), but several were expected,
            and none were found.
          - raises `ValueError` (i.e. does not return) any of the
            devices specifically requested by string isn't found.
        '''

        devs = {}
        
        for i in range(maxdev):
            logging.debug(f'Looking at HARP device')
            try:
                harp = MonsterHarp(i, self.hlib)
                
            except OSError as e:
                if e.errno == errno.EAGAIN:
                    continue
                
                raise
            
            if (harp.serial in serials) or \
               ("first" in serials) or \
               ("any" in serials) or \
               (len(serials) == 0):
                return harp

            if (harp.serial in serials) or ("all" in serials):
                devs[harp.serial] = harp
                continue

            # don't want it, ID doesn't match list
            del harp
            
        if ((len(serials) == 1) and (serials[0] != "all")) or (len(serials) == 0):
            # excatly one Harp requested, but none found.
            raise OSError(errno.ENXIO, "No Harp device attached")

        return devs
            
        

    @property
    def libraryVersion(self):
        '''
        Wrapper for `HH_GetLibraryVersion()`. Returns a string with the
        version of the binary library being used.
        '''
        buf = ctypes.create_string_buffer(b'', 128)
        self.hlib.HH_GetLibraryVersion(buf)
        return buf.value.decode('utf-8')

    
    @property
    def errorString(self):
        '''
        Wrapper for `HH_GetErrorString()`. Returns a string version
        of the specified error code. The string is essentially just
        the name of the error code itself, it presents zero value
        beyond what's already in PicoQuant's header files.
        '''
        return ErrorStringPortal(self.hlib)



# Path to the hhlib.so
monster_library = env.get('MINIQUANT_HHLIB', "hhlib.so")

# ctypes.CDLL binding to the library pointed to by
# the MINIQUANT_HARP_LIBRARY environment variable.
try:
    monster_binding = Monster(monster_library)
except:
    monster_binding = None



#!/usr/bin/python3

#
#        MiniQuant --  Wrapper for PicoQuant's HydraHarp access libraries.
#        Copyright (C) 2022 Florin Boariu.
#
#        This program is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        This program is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

## This implements a mock interface for the HydraHarp. The mock interface
## does not supply any meaningful data or behavior, it is only designed to
## not outright crash the application when run against.
##
## It should, however, be able to pass all unit tests that the real interface
## is also passing.
##
## The idea is that software based on miniquant can be run and tested using
## the mock interface in all instances that don't actually depend on reliable
## data. This should allow for automated unit testing of higher appliaction
## levels without actual access to a Harp device.

import numpy as np
import time

from miniquant.harp import MeasurementControl, MeasurementMode, EdgeSelection, \
    ReferenceClock, DeviceLimits, HarpWarnings, HarpFlags, HarpFeatures, \
    OrthogonalFlagsMap, HarpError

class ModuleInfo:
    def __init__(self, harp, modindex):
        self.harp = harp
        self.info = { "code": 3, "version": 4 }

class InputChannel:
    def __init__(self, harp, channel):
        self.harp = harp
        self.channel = channel
        self.moduleIndex = 1
        self.clearOnReadout = False
        self.enabled = False
        self.offset = 0
        self.cfd = {"level": 0.345, "zerox": 0.002 }
        self.countRate = 0
        self.histogram = np.array([0 for i in range(65536)])

    def __repr__(self):
        return "MockChannel[%d]" % self.channel
        

class MonsterHarp(object):

    def __repr__(self):
        return "MockHarp-%d" % self.serial
    
    def initialize(self, mode, clock):
        pass

    def calibrate(self):
        pass

    def clearHistogram(self):
        pass

    def acquire(self, aqtime):
        self._acq_stop_time = time.time()+aqtime
        
    def acqstop(self):
        self._acq_stop_time = time.time()

    @property
    def acquiring(self):
        '''
        This will mock aquisition by return `True` for a period of time
        after turning it on
        '''
        return time.time() < self._acq_stop_time

    @acquiring.setter
    def acquiring(self, enable):
        if enable:
            self.acquire(self.acquisitionTime)
        else:
            self.acqstop()

    # This is also trick, acquisition should turn itself off.

    @property
    def serialNumber(self):
        return self.serial

    @property
    def syncCfd(self):
        return self._syncCfd
    
    @syncCfd.setter
    def syncCfd(self, data):
        if isinstance(data, dict):
            self._syncCfd = data
        else:
            self._syncCfd = dict({"level": data[0], "zerox": data[1]})

    @property
    def histogramLength(self):
        if self._histogramLength is None:
            return None
        return (2**self._histogramLength) 

    @histogramLength.setter
    def histogramLength(self, val):
        if val > 6:
            raise HarpError("Should be <=6")
        self._histogramLength = val

    @property
    def resolution(self):
        return self.baseResolution[0] * 2**self.binning
    
    def __init__(self, id=0, hlib=None):
        self.serial = 4711
        self.hardwareInfo = { "model": 'MockHarp', "part": '23', "vesion": '17' }
        self.features = []
        self.channels = [ InputChannel(self, i) for i in range(4) ]
        self.modules =  [ ModuleInfo(self, i) for i in range(4) ]
        self.hardwareDebugInfo = "There is no harp."
        self.syncDiv = 1
        self._syncCfd = {"level": 0.312, "zerox": 0.011 }
        self.syncChannelOffset = 0.3
        self.stopOverflow = False
        self.binning = 0
        self.offset = 0
        self.baseResolution = (1e-12, DeviceLimits.MaxBinning)
        self._histogramLength = None
        
        self.measurementControl = {"control": MeasurementControl.ContC1Gated,
                                   "startedge": EdgeSelection.Rising,
                                   "stopedge": EdgeSelection.Falling }

        self.acquisitionTime = 1.0
        self._acq_stop_time = 0        
        self.clearOnAcquire = True
        self.syncRate = 23
        self.syncPeriod = 1/self.syncRate
        self.warnings = OrthogonalFlagsMap(HarpWarnings, code=0x801) # sync rate zeor, offset unnecessary
        self.flags = OrthogonalFlagsMap(HarpFlags, code=0x18) # ref lost, system error
        self.markers = [ None for i in range(4) ]
        self.markerHoldoffTime = 0.003
        self.fifo = np.array([0 for i in range(self.histogramLength or DeviceLimits.MaxHistogramLength)])
        

class Monster(object):

    def find(self, *serials, maxdev=8):

        r = []
        for s in serials:

            if s in [ "first", "any" ]:
                return MonsterHarp()

            if s in [ "all" ]:
                return [MonsterHarp()]

            if s in [ "unlikely" ]:
                raise OSError("Unit tests demand that this serial doesn't exist")            
            
            if isinstance(s, str) or type(s) == int:
                m = MonsterHarp()
                m.serial = s
                return m
            
            r.append(MonsterHarp())
            r[-1].serial = s
            
        return r

    def __new__(self, *args, **kwargs):
        # Make a singleton
        if not hasattr(self, 'instance'):
            self.instance = super(Monster, self).__new__(self)            
        return self.instance
    
    def __init__(self, library=None):
        self.libraryVersion = 'Mock v0.1'
        self.errorString = ''

monster_binding = Monster()



#!/usr/bin/pyton3

from os import environ as env

if 'MINIQUANT_TEST' in env and env['MINIQUANT_TEST'] != 0:
    from miniquant import mock_harp as harp
else:
    from miniquant import harp

import logging

class HarpManager(object):
    '''
    Manages a Miniquant Harp device and its configuration presets.
    '''

    def __init__(self, device='first', startup=None, settings=None):
        self.harpDevice = device
        self.harp = harp.Monster().find(self.harpDevice)
        
        if startup:
            self.initialize(startup)

        if settings:
            self.configure(settings)

        self._aquisitionTime = 1.0


    def initialize(self, startup):
        '''
        Initializes the harp and runs a calibration measurement.
        The `startup` parameter is expected to contain a configuration
        dictionary with the following startup settings:
        
          - `measurementMode`: one of the `miniquant.harp.MeasurementMonde` variables,
            as string (`Histogramming`, `T2`, `T3`, `Continuous`). The string is
            case sensitive, default is `Histogramming`.
        
          - `referenceClock`: one of the `miniquant.harp.ReferenceClock` values
            as string, i.e. `Internal` or `External`. Default is `Internal`.
        
        '''
        logging.info("Initializing harp, this will take a moment...")
        self.harp.initialize(getattr(harp.MeasurementMode, startup['measurementMode']),
                             getattr(harp.ReferenceClock, startup['referenceClock']))
        self.harp.calibrate()
        logging.info("Harp ready")


    def __obj_conf(self, device_object, config):
        '''
        '''
        for propKey in config:
            #logging.info("Property: %r" % propKey)
            try:
                #logging.info("Investigating %r.%s", device_object, propKey) 
                device_prop = getattr(device_object, propKey)
                if not propKey in [ 'channels' ]:#hasattr(device_prop, "__getitem__"):
                    # Handling a Harp property
                    logging.info ("%r.%s -> %r" % (device_object, propKey, config[propKey]))
                    setattr(device_object, propKey, config[propKey])
                else:
                    # Handling Channels
                    logging.info ("Channel: %r" % device_object)
                    for chdev,chconf in zip(device_prop, config[propKey]):
                        logging.info ("Descending into channel: %r" % chdev)
                        self.__obj_conf(device_object=chdev, config=chconf)
            except:
                logging.error("Failure setting property '%s' of %r" % (propKey, device_object))
                logging.error("Value is: %r" % (cfg[propKey]))
                raise
        


    def configure(self, settings):
        '''
        Configures the Harp using values from the dictionary `settings`.
        The dictionary may contain values for the `miniquant.harp.MonsterHarp`
        properties `syncDiv`, `syncCfd` (with `level` and `zerox` in a dictionary)
            `syncChannelOffset`, `histogramLength`, `binning`, `offset`
        
        It also accepts a `channels` key with an array of channel properties,
        containing `enabled`, `offset` and `cfd` (again with `level` and `zerox`).
        '''

        # __obj_conf() does the heavy lifting, this is essentially just
        # a user-safe frontend.
        #print ("Settings:", settings)
        self.__obj_conf(self.harp, settings)


    def logStatus(self):
        '''
        Rapid logging of current harp setup and status, for debugging and information.
        '''
        hdev = self.harp
        logging.info ("Histogram length: %r" % hdev.histogramLength)
        logging.info ("Base resolution: %r" % (hdev.baseResolution,))
        logging.info ("Real resolution: %r" % hdev.resolution)
        logging.info ("Histogram delta: %r" % (hdev.baseResolution[0]*(2**hdev.binning),))
        logging.info ("Histogram reach: %r" % (hdev.resolution*hdev.histogramLength,))
        logging.info ("Sync rate: %r" % hdev.syncRate)
        logging.info ("Sync period: %r" % hdev.syncPeriod)
        logging.info ("Count rates: %r" % [ch.countRate for ch in hdev.channels])
        logging.info ("Warnings: %r" % [ w for w in hdev.warnings] )
        logging.info ("Flags: %r" % [ f for f in hdev.flags] )

